package com.sencha.gxt.cell.core.client;

import com.google.gwt.core.client.GWT;
import com.google.gwt.safehtml.shared.SafeHtmlBuilder;
import com.google.gwt.user.client.Timer;
import com.sencha.gxt.core.client.Style;
import com.sencha.gxt.core.client.dom.XElement;

public class ProgressBarCell extends ResizeCell<Double> {

  public static interface ProgressBarAppearance {

    void render(SafeHtmlBuilder sb, Double value, int width);

    void updateText(XElement parent, String text);

  }

  protected ProgressBarAppearance appearance;

  private int increment = 10;
  private int interval = 300;
  private int duration = Style.DEFAULT;
  private Timer timer;

  private int i = 0;
  private boolean running;

  public ProgressBarCell() {
    this(GWT.<ProgressBarAppearance> create(ProgressBarAppearance.class));
  }

  public ProgressBarCell(ProgressBarAppearance appearance) {
    this.appearance = appearance;
  }

  /**
   * Initiates an auto-updating progress bar using the current duration,
   * increment, and interval.
   */
  public void auto(final Context context, final XElement parent) {
    if (timer == null) {
      timer = new Timer() {
        public void run() {
          int inc = getIncrement();
          updateProgress(context, parent, ((((i++ + inc) % inc) + 1) * (100 / inc)) * .01, null);
        }
      };
    }
    timer.scheduleRepeating(getInterval());
    running = true;
  }

  /**
   * Returns the duration.
   * 
   * @return the duration
   */
  public int getDuration() {
    return duration;
  }

  /**
   * Returns the bar's increment value.
   * 
   * @return the increment the increment
   */
  public int getIncrement() {
    return increment;
  }

  /**
   * Returns the bar's interval value.
   * 
   * @return the interval in milliseconds
   */
  public int getInterval() {
    return interval;
  }

  /**
   * Returns true if the progress bar is currently in a {@link #auto} operation.
   * 
   * @return true if waiting, else false
   */
  public boolean isRunning() {
    return running;
  }

  @Override
  public void render(Context context, Double value, SafeHtmlBuilder sb) {
    appearance.render(sb, value, getWidth());
  }

  /**
   * The length of time in milliseconds that the progress bar should run before
   * resetting itself (defaults to DEFAULT, in which case it will run
   * indefinitely until reset is called)
   * 
   * @param duration the duration in milliseconds
   */
  public void setDuration(int duration) {
    this.duration = duration;
  }

  /**
   * The number of progress update segments to display within the progress bar
   * (defaults to 10). If the bar reaches the end and is still updating, it will
   * automatically wrap back to the beginning.
   * 
   * @param increment the new increment
   */
  public void setIncrement(int increment) {
    this.increment = increment;
  }

  /**
   * Resets the progress bar value to 0 and text to empty string.
   */
  public void reset(Context context, XElement parent) {
    updateProgress(context, parent, 0, "");
    if (timer != null) {
      timer.cancel();
    }
    running = false;
  }

  /**
   * Sets the length of time in milliseconds between each progress update
   * (defaults to 300 ms).
   * 
   * @param interval the interval to set
   */
  public void setInterval(int interval) {
    this.interval = interval;
  }

  /**
   * Updates the progress bar value, and optionally its text. If the text
   * argument is not specified, any existing text value will be unchanged. To
   * blank out existing text, pass "". Note that even if the progress bar value
   * exceeds 1, it will never automatically reset -- you are responsible for
   * determining when the progress is complete and calling {@link #reset} to
   * clear and/or hide the control.
   * 
   * @param value A value between 0 and 1 (e.g., .5, defaults to 0)
   * @param text The string to display in the progress text element or null.
   */
  public void updateProgress(Context context, XElement parent, double value, String text) {
    value = Math.min(Math.max(value, 0), 1);

    setValue(context, parent, value);

    if (text != null) {
      updateText(parent, text);
    }
  }

  /**
   * Updates the progress bar text. If specified, textEl will be updated,
   * otherwise the progress bar itself will display the updated text.
   * 
   * @param text The string to display in the progress text element
   */
  public void updateText(XElement parent, String text) {
    appearance.updateText(parent, text);
  }

}
