package com.sencha.gxt.cell.core.client.form;

import java.util.HashSet;
import java.util.Set;

import com.google.gwt.cell.client.ValueUpdater;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.InputElement;
import com.google.gwt.dom.client.NativeEvent;
import com.sencha.gxt.cell.core.client.AbstractEventInputCell;
import com.sencha.gxt.cell.core.client.ResizableCell;
import com.sencha.gxt.cell.core.client.form.FieldCell.FieldViewData;
import com.sencha.gxt.core.client.dom.XElement;
import com.sencha.gxt.widget.core.client.event.BlurEvent;
import com.sencha.gxt.widget.core.client.event.FocusEvent;

public abstract class FieldCell<T> extends AbstractEventInputCell<T, FieldViewData> implements ResizableCell {

  public static interface FieldAppearance {

    void onEmpty(Element parent, boolean empty);

    void onFocus(Element parent, boolean focus);

    void onValid(Element parent, boolean valid);

    void setReadOnly(Element parent, boolean readonly);

  }

  public static class FieldAppearanceOptions {
    private int width;
    private int height;
    private boolean readonly;
    private String emptyText;
    private String name = "";
    private boolean hideTrigger;
    private boolean editable = true;

    public FieldAppearanceOptions() {

    }

    public FieldAppearanceOptions(int width, int height, boolean readonly) {
      this.readonly = readonly;
      this.width = width;
      this.height = height;
    }

    public FieldAppearanceOptions(int width, int height, boolean readonly, String emptyText) {
      this(width, height, readonly);
      this.emptyText = emptyText;
    }

    public String getEmptyText() {
      return emptyText;
    }

    public int getHeight() {
      return height;
    }

    public String getName() {
      return name;
    }

    public int getWidth() {
      return width;
    }

    public boolean isEditable() {
      return editable;
    }

    public boolean isHideTrigger() {
      return hideTrigger;
    }

    public boolean isReadonly() {
      return readonly;
    }

    public void setEditable(boolean editable) {
      this.editable = editable;
    }

    public void setEmptyText(String emptyText) {
      this.emptyText = emptyText;
    }

    public void setHeight(int height) {
      this.height = height;
    }

    public void setHideTrigger(boolean hideTrigger) {
      this.hideTrigger = hideTrigger;
    }

    public void setName(String name) {
      this.name = name;
    }

    public void setReadonly(boolean readonly) {
      this.readonly = readonly;
    }

    public void setWidth(int width) {
      this.width = width;
    }

  }

  public static class FieldViewData extends ViewData {

    private Object focusValue;

    private boolean hasFocus;

    private boolean mimicing;

    public FieldViewData(String value) {
      super(value);
    }

    public Object getFocusValue() {
      return focusValue;
    }

    public boolean isHasFocus() {
      return hasFocus;
    }

    public boolean isMimicing() {
      return mimicing;
    }

    public void setFocusValue(Object focusValue) {
      this.focusValue = focusValue;
    }

    public void setHasFocus(boolean hasFocus) {
      this.hasFocus = hasFocus;
    }

    public void setMimicing(boolean mimicing) {
      this.mimicing = mimicing;
    }

  }

  /**
   * Get the events consumed by the input cell.
   * 
   * @param userEvents the events consumed by the subclass
   * @return the events
   */
  private static Set<String> getConsumedEventsImpl(Set<String> userEvents) {
    Set<String> events = new HashSet<String>();
    events.add("click");
    events.add("keyup");
    events.add("keypress");
    events.add("focus");
    events.add("blur");
    events.add("mouseover");
    events.add("mouseout");
    events.add("mousedown");
    events.add("mouseup");
    events.add("change");

    if (userEvents != null && userEvents.size() > 0) {
      events.addAll(userEvents);
    }
    return events;
  }

  /**
   * Get the events consumed by the input cell.
   * 
   * @param userEvents the events consumed by the subclass
   * @return the events
   */
  private static Set<String> getConsumedEventsImpl(String... consumedEvents) {
    Set<String> userEvents = new HashSet<String>();
    if (consumedEvents != null) {
      for (String event : consumedEvents) {
        userEvents.add(event);
      }
    }
    return getConsumedEventsImpl(userEvents);
  }

  protected final FieldAppearance appearance;
  protected int height = -1;
  protected Context lastContext;
  protected XElement lastParent;
  protected T lastValue;
  protected ValueUpdater<T> lastValueUpdater;
  protected int width = -1;
  private boolean readOnly;

  public FieldCell(FieldAppearance appearance, Set<String> consumedEvents) {
    super(getConsumedEventsImpl(consumedEvents));
    this.appearance = appearance;
  }

  public FieldCell(FieldAppearance appearance, String... consumedEvents) {
    super(getConsumedEventsImpl(consumedEvents));
    this.appearance = appearance;
  }

  public int getHeight() {
    return height;
  }

  public int getWidth() {
    return width;
  }

  public boolean isReadOnly() {
    return readOnly;
  }

  public void onBrowserEvent(Context context, Element parent, T value, NativeEvent event, ValueUpdater<T> valueUpdater) {
    XElement x = parent.cast();
    String eventType = event.getType();
    if ("keydown".equals(eventType)) {
      saveContext(context, parent, event, valueUpdater, value);
      onKeyDown(context, parent, value, event, valueUpdater);
    } else if ("keyup".equals(eventType)) {
      onKeyUp(context, parent, value, event, valueUpdater);
    } else if ("keypress".equals(eventType)) {
      onKeyPress(parent, event);
    } else if ("blur".equals(eventType)) {
      saveContext(context, parent, event, valueUpdater, value);
      onBlur(context, x, value, event);
    } else if ("focus".equals(eventType)) {
      saveContext(context, parent, event, valueUpdater, value);
      onFocus(context, x, value, event);
    } else if ("mouseover".equals(eventType)) {
      onMouseOver(x, event);
    } else if ("mouseout".equals(eventType)) {
      onMouseOut(x, event);
    } else if ("mousedown".equals(eventType)) {
      saveContext(context, parent, event, valueUpdater, value);
      onMouseDown(x, event);
    } else if ("click".equals(eventType)) {
      saveContext(context, parent, event, valueUpdater, value);
      onClick(x, event);
    }
  }

  public abstract void onEmpty(XElement parent, boolean empty);

  public void onValid(XElement parent, boolean valid) {
    appearance.onValid(parent, valid);
  }

  public void setHeight(int height) {
    this.height = height;
  }

  public void setName(XElement parent, String name) {
    getInputElement(parent).setAttribute("name", name);
  }

  public void setReadOnly(boolean readOnly) {
    this.readOnly = readOnly;
  }

  @Override
  public void setSize(int width, int height) {
    this.width = width;
    this.height = height;
  }

  public void setWidth(int width) {
    this.width = width;
  }

  protected FieldViewData checkViewData(Context context, String value) {
    Object key = context.getKey();
    FieldViewData viewData = getViewData(key);
    if (viewData != null && viewData.getCurrentValue().equals(value)) {
      clearViewData(key);
      viewData = null;
    }
    return viewData;
  }

  protected void clearContext() {
    lastContext = null;
    lastParent = null;
    lastValueUpdater = null;
    lastValue = null;
  }

  protected FieldViewData ensureViewData(Context context, XElement parent) {
    Object key = context.getKey();
    FieldViewData viewData = getViewData(key);
    if (viewData == null) {
      viewData = new FieldViewData(((InputElement) getInputElement(parent).cast()).getValue());
      setViewData(key, viewData);
    }
    return viewData;
  }

  @Override
  protected void finishEditing(Element parent, T value, Object key, ValueUpdater<T> valueUpdater) {
    super.finishEditing(parent, value, key, valueUpdater);
    clearViewData(key);
  }

  protected boolean hasFocus(Context context, XElement parent) {
    return ensureViewData(context, parent).isHasFocus();
  }

  protected void onBlur(Context context, XElement parent, T value, NativeEvent event) {
    FieldViewData viewData = ensureViewData(context, parent);
    if (viewData.isHasFocus()) {
      appearance.onFocus(parent, false);
      viewData.setHasFocus(false);
      if (viewData.getFocusValue() != value) {
        // TODO and updates to value provider will fire change event, should we
        // fire here?
      }
      fireEvent(context, new BlurEvent());
    }
  }

  protected void onClick(XElement x, NativeEvent event) {

  }

  protected void onFocus(Context context, XElement parent, T value, NativeEvent event) {
    FieldViewData viewData = ensureViewData(context, parent);
    if (!viewData.isHasFocus()) {
      viewData.setHasFocus(true);
      viewData.setFocusValue(value);
      appearance.onFocus(parent, true);
      fireEvent(context, new FocusEvent());
    }
  }

  protected void onKeyDown(Context context, Element parent, T value, NativeEvent event, ValueUpdater<T> valueUpdater) {

  }

  protected void onKeyPress(Element parent, NativeEvent event) {

  }

  protected void onKeyUp(Context context, Element parent, T value, NativeEvent event, ValueUpdater<T> valueUpdater) {

  }

  protected void onMouseDown(XElement parent, NativeEvent event) {

  }

  protected void onMouseOut(XElement parent, NativeEvent event) {

  }

  protected void onMouseOver(XElement parent, NativeEvent event) {

  }

  protected void onMouseUp(XElement parent, NativeEvent event) {

  }

  protected void saveContext(Context context, Element parent, NativeEvent event, ValueUpdater<T> valueUpdater, T value) {
    lastContext = context;
    lastParent = parent.cast();
    lastValueUpdater = valueUpdater;
    lastValue = value;
  }
}
