package com.sencha.gxt.cell.core.client.form;

import java.text.ParseException;

import com.google.gwt.cell.client.ValueUpdater;
import com.google.gwt.core.client.GWT;
import com.google.gwt.core.client.Scheduler;
import com.google.gwt.core.client.Scheduler.ScheduledCommand;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.InputElement;
import com.google.gwt.dom.client.NativeEvent;
import com.google.gwt.event.dom.client.KeyCodes;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.safehtml.shared.SafeHtmlBuilder;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.Event.NativePreviewEvent;
import com.google.gwt.user.client.Event.NativePreviewHandler;
import com.sencha.gxt.core.client.dom.XElement;
import com.sencha.gxt.widget.core.client.event.ParseErrorEvent;
import com.sencha.gxt.widget.core.client.event.TriggerClickEvent;

public class TriggerFieldCell<T> extends ValueBaseFieldCell<T> {

  public interface TriggerFieldAppearance extends ValueBaseFieldAppearance {
    void onTriggerClick(XElement parent, boolean click);

    void onTriggerOver(XElement parent, boolean over);

    void render(SafeHtmlBuilder sb, String value, FieldAppearanceOptions options);

    void setTriggerVisible(XElement parent, boolean visible);

    void setEditable(XElement parent, boolean editable);

    boolean triggerIsOrHasChild(XElement parent, Element target);
  }

  private class MouseDownHandler implements NativePreviewHandler {

    private Context context;
    private final Element parent;
    private final HandlerRegistration reg;

    public MouseDownHandler(Context context, Element parent) {
      this.context = context;
      this.parent = parent;
      this.reg = Event.addNativePreviewHandler(this);
    }

    public void onPreviewNativeEvent(NativePreviewEvent event) {
      if ("mousedown".equals(event.getNativeEvent().getType())) {
        XElement target = event.getNativeEvent().getEventTarget().cast();

        if (!isFocusClick(parent.<XElement> cast(), target)) {
          reg.removeHandler();
          mimicBlur(context, parent.<XElement> cast(), event, lastValue);
        }
      }
    }
  }

  protected final TriggerFieldAppearance appearance;

  private boolean editable = true;
  private boolean hideTrigger;

  public TriggerFieldCell() {
    this(GWT.<TriggerFieldAppearance> create(TriggerFieldAppearance.class));
  }

  public TriggerFieldCell(TriggerFieldAppearance appearance) {
    super(appearance);
    this.appearance = appearance;
  }

  /**
   * Returns true if the field is editable.
   * 
   * @return true if editable
   */
  public boolean isEditable() {
    return editable;
  }

  public boolean isHideTrigger(Context context) {
    return hideTrigger;
  }

  public void onBrowserEvent(Context context, Element parent, T value, NativeEvent event, ValueUpdater<T> valueUpdater) {
    super.onBrowserEvent(context, parent, value, event, valueUpdater);
    String eventType = event.getType();
    Object key = context.getKey();
    if ("click".equals(eventType)) {
      onClick(context, parent.<XElement> cast(), event, value, valueUpdater);
    } else if ("keyup".equals(eventType)) {
      // Record keys as they are typed.
      FieldViewData vd = getViewData(key);
      if (vd == null) {
        vd = new FieldViewData(getText(XElement.as(parent)));
        setViewData(key, vd);
      }
      vd.setCurrentValue(getText(XElement.as(parent)));
    }
  }

  @Override
  public void render(Context context, T value, SafeHtmlBuilder sb) {
    String v = "";
    if (value != null) {
      v = getPropertyEditor().render(value);
    }

    FieldViewData viewData = checkViewData(context, v);

    FieldAppearanceOptions options = new FieldAppearanceOptions(width, height, isReadOnly());
    options.setEmptyText(getEmptyText());
    options.setHideTrigger(isHideTrigger(context));
    options.setEditable(editable);

    String s = (viewData != null) ? viewData.getCurrentValue() : v;
    appearance.render(sb, s == null ? "" : s, options);

  }

  public void setEditable(XElement parent, boolean editable) {
    this.editable = editable;

    appearance.setEditable(parent, editable);

    InputElement fromEl = appearance.getInputElement(parent).cast();
    if (!isReadOnly()) {
      fromEl.setPropertyBoolean("readOnly", !editable);
    }
  }

  public void setHideTrigger(boolean hideTrigger) {
    this.hideTrigger = hideTrigger;
  }

  @Override
  protected void finishEditing(Element parent, T value, Object key, ValueUpdater<T> valueUpdater) {
    String newValue = getText(XElement.as(parent));
    
    // Get the view data.
    FieldViewData vd = getViewData(key);
    if (vd == null) {
      vd = new FieldViewData(newValue);
      setViewData(key, vd);
    }
    vd.setCurrentValue(newValue);

    // Fire the value updater if the value has changed.
    if (valueUpdater != null && !vd.getCurrentValue().equals(vd.getLastValue())) {
      vd.setLastValue(newValue);
      try {
        valueUpdater.update(newValue.equals("") ? null : getPropertyEditor().parse(newValue));
      } catch (ParseException e) {
        vd.setCurrentValue("");
        valueUpdater.update(null);
        setText(parent.<XElement> cast(), "");
        fireEvent(new ParseErrorEvent(e));
      }
    }

    // Blur the element.
    super.finishEditing(parent, value, key, valueUpdater);
  }

  protected boolean isFocusClick(XElement parent, XElement target) {
    return parent.isOrHasChild(target);
  }

  protected void mimicBlur(Context context, XElement parent, NativePreviewEvent pe, T value) {
    Element target = pe.getNativeEvent().getEventTarget().cast();
    if (!parent.isOrHasChild(target) && validateBlur(target)) {
      triggerBlur(context, parent, value);
    }
  }

  @Override
  protected void onBlur(Context context, XElement parent, T value, NativeEvent event) {
    // do nothing
  }

  protected void onClick(Context context, XElement parent, NativeEvent event, T value, ValueUpdater<T> updater) {
    Element target = event.getEventTarget().cast();

    if (!isReadOnly() && !isReadOnly() && appearance.triggerIsOrHasChild(parent.<XElement> cast(), target)) {
      onTriggerClick(context, parent, event, value, updater);
    }

    if (!editable) {
      event.preventDefault();
      event.stopPropagation();
    }
  }

  @Override
  protected void onFocus(Context context, XElement parent, T value, NativeEvent event) {
    super.onFocus(context, parent, value, event);
    FieldViewData view = ensureViewData(context, parent);
    if (!view.isMimicing()) {
      appearance.onFocus(parent, true);
      view.setMimicing(true);
      new MouseDownHandler(context, parent);
    }
  }

  @Override
  protected void onKeyDown(final Context context, final Element parent, final T value, NativeEvent event,
      ValueUpdater<T> valueUpdater) {
    super.onKeyDown(context, parent, value, event, valueUpdater);
    if (event.getKeyCode() == KeyCodes.KEY_TAB) {
      Scheduler.get().scheduleDeferred(new ScheduledCommand() {
        @Override
        public void execute() {
          if (hasFocus(context, XElement.as(parent))) {
            triggerBlur(context, parent.<XElement> cast(), value);
          }
        }
      });
    }
  }

  @Override
  protected void onMouseOut(XElement parent, NativeEvent event) {
    super.onMouseOut(parent, event);
    XElement target = event.getEventTarget().cast();
    if (appearance.triggerIsOrHasChild(parent.<XElement> cast(), target)) {
      appearance.onTriggerOver(parent.<XElement> cast(), false);
    }
  }

  @Override
  protected void onMouseOver(XElement parent, NativeEvent event) {
    super.onMouseOver(parent, event);
    XElement target = event.getEventTarget().cast();
    if (appearance.triggerIsOrHasChild(parent.<XElement> cast(), target)) {
      appearance.onTriggerOver(parent.<XElement> cast(), true);
    }
  }

  protected void onTriggerClick(Context context, XElement parent, NativeEvent event, T value, ValueUpdater<T> updater) {
    fireEvent(context, new TriggerClickEvent());
  }

  protected void triggerBlur(Context context, final XElement parent, T value) {
    Scheduler.get().scheduleDeferred(new ScheduledCommand() {
      @Override
      public void execute() {
        getInputElement(parent).blur();
      }
    });

    ensureViewData(context, parent).setMimicing(false);
    super.onBlur(context, parent, value, null);
    finishEditing(parent, value, context.getKey(), lastValueUpdater);
  }

  protected boolean validateBlur(Element target) {
    return true;
  }

}
