package com.sencha.gxt.widget.core.client.form;

import java.util.ArrayList;
import java.util.List;

import com.google.gwt.core.client.Scheduler;
import com.google.gwt.core.client.Scheduler.ScheduledCommand;
import com.google.gwt.editor.client.EditorDelegate;
import com.google.gwt.editor.client.EditorError;
import com.google.gwt.editor.client.HasEditorDelegate;
import com.google.gwt.editor.client.HasEditorErrors;
import com.google.gwt.editor.client.LeafValueEditor;
import com.google.gwt.event.logical.shared.HasValueChangeHandlers;
import com.google.gwt.event.logical.shared.ValueChangeEvent;
import com.google.gwt.event.logical.shared.ValueChangeHandler;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.Event;
import com.google.gwt.user.client.ui.HasName;
import com.google.gwt.user.client.ui.HasValue;
import com.sencha.gxt.cell.core.client.form.FieldCell;
import com.sencha.gxt.core.client.util.DelayedTask;
import com.sencha.gxt.core.client.util.Util;
import com.sencha.gxt.widget.core.client.cell.CellComponent;
import com.sencha.gxt.widget.core.client.event.BlurEvent;
import com.sencha.gxt.widget.core.client.event.BlurEvent.BlurHandler;
import com.sencha.gxt.widget.core.client.event.FocusEvent;
import com.sencha.gxt.widget.core.client.event.FocusEvent.FocusHandler;
import com.sencha.gxt.widget.core.client.event.InvalidEvent;
import com.sencha.gxt.widget.core.client.event.InvalidEvent.HasInvalidHandlers;
import com.sencha.gxt.widget.core.client.event.InvalidEvent.InvalidHandler;
import com.sencha.gxt.widget.core.client.event.ValidEvent;
import com.sencha.gxt.widget.core.client.event.ValidEvent.HasValidHandlers;
import com.sencha.gxt.widget.core.client.event.ValidEvent.ValidHandler;
import com.sencha.gxt.widget.core.client.form.error.DefaultEditorError;
import com.sencha.gxt.widget.core.client.form.error.ElementErrorHandler;
import com.sencha.gxt.widget.core.client.form.error.ErrorHandler;
import com.sencha.gxt.widget.core.client.form.error.SideErrorHandler;
import com.sencha.gxt.widget.core.client.form.error.TitleErrorHandler;
import com.sencha.gxt.widget.core.client.form.error.ToolTipErrorHandler;

public abstract class Field<T> extends CellComponent<T> implements IsField<T>, HasValue<T>, HasValueChangeHandlers<T>,
    HasName, HasInvalidHandlers, HasValidHandlers, LeafValueEditor<T>, HasEditorErrors<T>, HasEditorDelegate<T> {

  protected String forceInvalidText;
  protected DelayedTask validationTask;
  protected boolean preventMark;
  protected T originalValue;
  protected boolean hasFocus;

  private boolean autoValidate;
  private List<Validator<T>> validators = new ArrayList<Validator<T>>();
  private int validationDelay = 250;
  private String name;
  private boolean validateOnBlur = true;
  private EditorDelegate<T> delegate;
  private ErrorHandler errorSupport;
  private boolean valid = true;

  protected Field(FieldCell<T> cell) {
    super(cell, false);
    addBlurHandler(new BlurHandler() {

      @Override
      public void onBlur(BlurEvent event) {
        onLogicalBlur(event);
      }
    });
    addFocusHandler(new FocusHandler() {

      @Override
      public void onFocus(FocusEvent event) {
        onLogicalFocus(event);
      }
    });

    setErrorSupport(new SideErrorHandler(this));
  }

  @Override
  public HandlerRegistration addInvalidHandler(InvalidHandler handler) {
    return addHandler(handler, InvalidEvent.getType());
  }

  public void addValidator(Validator<T> validator) {
    validators.add(validator);
  }

  @Override
  public HandlerRegistration addValidHandler(ValidHandler handler) {
    return addHandler(handler, ValidEvent.getType());
  }

  @Override
  public HandlerRegistration addValueChangeHandler(ValueChangeHandler<T> handler) {
    return addHandler(handler, ValueChangeEvent.getType());
  }

  /**
   * Clears the value from the field.
   */
  public void clear() {
    boolean restore = preventMark;
    preventMark = true;
    setValue(null, false);
    preventMark = restore;
    clearInvalid();
  }

  /**
   * Clear any invalid styles / messages for this field.
   */
  public void clearInvalid() {
    valid = true;
    getCell().onValid(getElement(), true);
    forceInvalidText = null;
    errorSupport.clearInvalid();
    fireEvent(new ValidEvent());
  }

  /**
   * Forces the field to be invalid using the given error message. When using
   * this feature, {@link #clearInvalid()} must be called to clear the error.
   * Also, no other validation logic will execute.
   * 
   * @param msg the error text
   */
  public void forceInvalid(String msg) {
    forceInvalidText = msg;
    markInvalid(msg);
  }

  @Override
  public FieldCell<T> getCell() {
    return (FieldCell<T>) super.getCell();
  }

  /**
   * Returns the field's error support instance.
   * 
   * @return the error support
   */
  public ErrorHandler getErrorSupport() {
    return errorSupport;
  }

  /**
   * Gets the widget's name.
   * 
   * @return the widget's new name
   */
  @Override
  public String getName() {
    return name;
  }

  /**
   * Returns true if the value is validate on blur.
   * 
   * @return the validate on blur state
   */
  public boolean getValidateOnBlur() {
    return validateOnBlur;
  }

  /**
   * Returns the field's validation delay in milliseconds.
   * 
   * @return the delay in milliseconds
   */
  public int getValidationDelay() {
    return validationDelay;
  }

  /**
   * Returns the field's validators.
   * 
   * @return the validators
   */
  public List<Validator<T>> getValidators() {
    return validators;
  }

  /**
   * Returns true if the field value is validated on each key press.
   * 
   * @return the auto validate state
   */
  public boolean isAutoValidate() {
    return autoValidate;
  }

  /**
   * Returns the read only state.
   * 
   * @return <code>true</code> if read only, otherwise <code>false</code>
   */
  public boolean isReadOnly() {
    return getCell().isReadOnly();
  }

  /**
   * Returns whether or not the field value is currently valid.
   * 
   * @return true if valid
   */
  public boolean isValid() {
    return isValid(false);
  }

  @Override
  public boolean isValid(boolean preventMark) {
    if (disabled) {
      return true;
    }
    boolean restore = this.preventMark;
    this.preventMark = preventMark;
    boolean result = validateValue(getValue());
    if (result) {
      // activeErrorMessage = null;
    }
    this.preventMark = restore;
    return result;
  }

  /**
   * Marks this field as invalid. Validation will still run if called again, and
   * the error message will be changed or cleared based on validation. To set a
   * error message that will not be cleared until manually cleared see
   * {@link #forceInvalid(String)}
   * 
   * Calling this will also register an error in the editor, if any.
   * 
   * @param msg the validation message
   */
  @SuppressWarnings("unchecked")
  public void markInvalid(String msg) {
    if (delegate != null) {
      delegate.recordError(msg, getValue(), null);
    }
    markInvalid(Util.createList(new DefaultEditorError(this, msg, getValue())));
  }

  @Override
  public void onBrowserEvent(Event event) {
    super.onBrowserEvent(event);
    switch (event.getTypeInt()) {
      case Event.ONKEYUP:
        onKeyUp(event);
        break;
      case Event.ONKEYDOWN:
        onKeyDown(event);
        break;
      case Event.ONKEYPRESS:
        onKeyPress(event);
        break;
    }
  }

  @Override
  public void redraw() {
    super.redraw();
    if (!valid) {
      getCell().onValid(getElement(), false);
    }
  }

  public void removeValidator(Validator<T> validator) {
    validators.remove(validator);
  }

  @Override
  public void reset() {
    boolean restore = preventMark;
    preventMark = true;
    setValue(originalValue, false);
    preventMark = restore;
    clearInvalid();
  }

  /**
   * Sets whether the value is validated on each key press (defaults to false).
   * 
   * @param autoValidate true to validate on each key press
   */
  public void setAutoValidate(boolean autoValidate) {
    if (!this.autoValidate && autoValidate) {
      validationTask = new DelayedTask() {
        @Override
        public void onExecute() {
          validate();
        }
      };
    } else if (!autoValidate && validationTask != null) {
      validationTask.cancel();
      validationTask = null;
    }
    this.autoValidate = autoValidate;
  }

  @Override
  public void setDelegate(EditorDelegate<T> delegate) {
    this.delegate = delegate;
  }

  /**
   * Sets the field's error support instance.
   * 
   * <p />
   * See the following concrete implementations:
   * <ul>
   * <li>{@link SideErrorHandler}</li>
   * <li>{@link ToolTipErrorHandler}</li>
   * <li>{@link TitleErrorHandler}</li>
   * <li>{@link ElementErrorHandler}</li>
   * </ul>
   * 
   * @param error the error support
   */
  public void setErrorSupport(ErrorHandler error) {
    this.errorSupport = error;
  }

  @Override
  public void setName(String name) {
    this.name = name;
    getCell().setName(getElement(), name);
  }

  public void setOriginalValue(T value) {
    originalValue = value;
  }

  /**
   * Sets the field's read only state.
   * 
   * @param readOnly the read only state
   */
  public void setReadOnly(boolean readOnly) {
    getCell().setReadOnly(readOnly);
  }

  /**
   * Sets whether the field should validate when it loses focus (defaults to
   * true).
   * 
   * @param validateOnBlur true to validate on blur, otherwise false
   */
  public void setValidateOnBlur(boolean validateOnBlur) {
    this.validateOnBlur = validateOnBlur;
  }

  /**
   * Sets length of time in milliseconds after user input begins until
   * validation is initiated (defaults to 250).
   * 
   * @param validationDelay the delay in milliseconds
   */
  public void setValidationDelay(int validationDelay) {
    this.validationDelay = validationDelay;
  }

  @Override
  public void showErrors(List<EditorError> errors) {
    // all errors should apply to this element only

    for (EditorError error : errors) {
      assert error.getEditor() == this;
      error.setConsumed(true);
    }

    if (errors.size() > 0) {
      errorSupport.markInvalid(errors);
    } else {
      clearInvalid();
    }
  }

  /**
   * Validates the field value.
   * 
   * @return <code>true</code> if valid, otherwise <code>false</code>
   */
  public boolean validate() {
    return validate(false);
  }

  /**
   * Validates the field value.
   * 
   * @param preventMark true to not mark the field valid and fire invalid event
   *          when invalid
   * @return true if valid, otherwise false
   */
  public boolean validate(boolean preventMark) {
    if (disabled) {
      clearInvalid();
      return true;
    }

    boolean restore = this.preventMark;
    this.preventMark = preventMark;
    boolean result = validateValue(getValue());
    this.preventMark = restore;
    if (result) {
      clearInvalid();
      // activeErrorMessage = null;
    }
    return result;
  }

  /**
   * Actual implementation of markInvalid, which bypasses recording an error in
   * the editor peer.
   * 
   * @param msg the validation message
   */
  protected void markInvalid(List<EditorError> msg) {
    if (preventMark) {
      return;
    }
    valid = false;
    getCell().onValid(getElement(), false);

    errorSupport.markInvalid(msg);

    fireEvent(new InvalidEvent(msg));
  }

  protected void onKeyDown(Event event) {

  }

  protected void onKeyPress(Event event) {

  }

  protected void onKeyUp(Event event) {
    if (autoValidate && validationTask != null) {
      validationTask.delay(validationDelay);
    }
  }

  protected void onLogicalBlur(BlurEvent event) {
    if (hasFocus) {
      hasFocus = false;
      if (validateOnBlur) {
        Scheduler.get().scheduleEntry(new ScheduledCommand() {
          @Override
          public void execute() {
            validate();
          }
        });
      }
    }
  }

  protected void onLogicalFocus(FocusEvent event) {
    if (!hasFocus) {
      hasFocus = true;
    }
  }

  protected boolean validateValue(T value) {
    if (forceInvalidText != null) {
      markInvalid(forceInvalidText);
      valid = false;
      return false;
    }
    List<EditorError> errors = new ArrayList<EditorError>();
    for (int i = 0; i < validators.size(); i++) {
      List<EditorError> temp = validators.get(i).validate(this, value);
      if (temp != null && temp.size() > 0) {
        errors.addAll(temp);
      }
    }
    if (errors.size() > 0) {
      markInvalid(errors);
      valid = false;
      return false;
    }
    valid = true;
    return true;
  }
}
